/*
 * basiclu_get_factors.c
 *
 * Copyright (C) 2016-2017  Lukas Schork
 *
 */

#include "lu_internal.h"

lu_int basiclu_get_factors
(
    lu_int istore[],
    double xstore[],
    lu_int Li[],
    double Lx[],
    lu_int Ui[],
    double Ux[],
    lu_int Wi[],
    double Wx[],
    lu_int rowperm[],
    lu_int colperm[],
    lu_int Lcolptr[],
    lu_int Lrowidx[],
    double Lvalue_[],
    lu_int Ucolptr[],
    lu_int Urowidx[],
    double Uvalue_[]
)
{
    struct lu this;
    lu_int m, status;

    status = lu_load(&this, istore, xstore, Li, Lx, Ui, Ux, Wi, Wx);
    if (status != BASICLU_OK)
        return status;
    if (this.nupdate != 0)
    {
        status = BASICLU_ERROR_invalid_call;
        return lu_save(&this, istore, xstore, status);
    }
    m = this.m;

    if (rowperm)
        memcpy(rowperm, this.pivotrow, m*sizeof(lu_int));
    if (colperm)
        memcpy(colperm, this.pivotcol, m*sizeof(lu_int));

    if (Lcolptr && Lrowidx && Lvalue_)
    {
        const lu_int *Lbegin_p  = this.Lbegin_p;
        const lu_int *Lindex    = this.Lindex;
        const double *Lvalue    = this.Lvalue;
        const lu_int *p         = this.p;
        lu_int *pinv            = this.iwork1; /* size m workspace */
        lu_int i, ipivot, k, put, pos;

        /*
         * Fill L columnwise including the unit elements at the front of each
         * column. L[:,k] holds the elimination factors from the k-th pivot
         * step. Row indices are filled unpermuted and permuted later.
         * Build inverse row permutation.
         */
        #ifndef NDEBUG
        for (k = 0; k < m; k++)
            pinv[k] = -1;
        #endif

        put = 0;
        for (k = 0; k < m; k++)
        {
            ipivot = p[k];
            Lcolptr[k] = put;
            Lrowidx[put] = ipivot;
            Lvalue_[put++] = 1.0;
            for (pos = Lbegin_p[k]; (i = Lindex[pos]) >= 0; pos++)
            {
                Lrowidx[put] = i;
                Lvalue_[put++] = Lvalue[pos];
            }
            pinv[ipivot] = k;
        }
        Lcolptr[m] = put;
        assert(put == this.Lnz+m);

        #ifndef NDEBUG
        for (k = 0; k < m; k++)
            assert(pinv[k] >= 0);
        #endif

        /* permute row indices */
        for (pos = put-1; pos >= 0; pos--)
        {
            Lrowidx[pos] = pinv[Lrowidx[pos]];
        }
    }

    if (Ucolptr && Urowidx && Uvalue_)
    {
        const lu_int *Ubegin    = this.Ubegin;
        const lu_int *Uindex    = this.Uindex;
        const double *Uvalue    = this.Uvalue;
        const double *row_pivot = this.row_pivot;
        const lu_int *pivotrow  = this.pivotrow;
        lu_int *pinv            = this.iwork1; /* size m workspace */
        lu_int i, ipivot, k, put, pos;

        /*
         * Fill U columnwise with the pivot element at the end of each column.
         * U[:,k] holds the column of B from the k-th pivot step.
         * Row indices are filled unpermuted and permuted later.
         * Build inverse row permutation.
         */
        #ifndef NDEBUG
        for (k = 0; k < m; k++)
            pinv[k] = -1;
        #endif

        put = 0;
        for (k = 0; k < m; k++)
        {
            ipivot = pivotrow[k];
            Ucolptr[k] = put;
            for (pos = Ubegin[ipivot]; (i = Uindex[pos]) >= 0; pos++)
            {
                Urowidx[put] = i;
                Uvalue_[put++] = Uvalue[pos];
            }
            Urowidx[put] = ipivot;
            Uvalue_[put++] = row_pivot[ipivot];
            pinv[ipivot] = k;
        }
        Ucolptr[m] = put;
        assert(put == this.Unz+m);

        #ifndef NDEBUG
        for (k = 0; k < m; k++)
            assert(pinv[k] >= 0);
        #endif

        /* permute row indices */
        for (pos = put-1; pos >= 0; pos--)
        {
            Urowidx[pos] = pinv[Urowidx[pos]];
        }
    }

    return BASICLU_OK;
}
